from .chapterdownload import download_chapter
from .errors import ManganeloError, RequestError
from .hometooltips import get_home_page
from .models import Chapter, HomeStoryTooltip, SearchResult, StoryPage
from .storypage import get_story_page
from .storysearch import get_search_results


def get_chapter_list(url: str) -> list[Chapter]:
	return get_story_page(url).chapter_list



import ast
import datetime as dt
import re

from pydantic import BaseModel

from ..chapterdownload import download_chapter
from ..common import utils
from typing import Union


class Chapter(BaseModel):
    title: str
    url: str
    chapter: Union[int, float]
    views: int
    uploaded: dt.datetime

    def download(self, path):
        return download_chapter(self.url, path)

    @staticmethod
    def from_soup(soup):
        obj = Chapter(
            title=_parse_title(soup),
            url=_parse_url(soup),
            chapter=-1,
            views=_parse_views(soup),
            uploaded=_parse_uploaded(soup)
        )

        obj.chapter = _parse_chapter(obj.url)

        return obj


def _parse_title(soup):
    return soup.find("a").text


def _parse_url(soup):
    return soup.find("a").get("href")


def _parse_chapter(url: str) -> Union[int, float]:
    return ast.literal_eval(re.split("[-_]", url.split("chapter")[-1])[-1])


def _parse_views(soup):
    s = soup.find_all("span", class_="chapter-view text-nowrap")[-1].text.replace(",", "")
    return utils.parse_views(s)


def _parse_uploaded(soup):
    s = soup.find("span", class_="chapter-time text-nowrap").get("title")
    return utils.parse_date(s, "%b %d,%Y %H:%M")



import collections
import threading
import time
import urllib.parse

import requests

from manganelo.common.constants import ROOT_URL


class _RateLimiter:
    def __init__(self, max_calls, period):
        self.calls = collections.deque()

        self.period = period
        self.max_calls = max_calls
        self.lock = threading.Lock()

    def __enter__(self):
        with self.lock:
            if len(self.calls) >= self.max_calls:
                until = time.time() + self.period - self._timespan
                sleeptime = until - time.time()
                if sleeptime > 0:
                    time.sleep(sleeptime)
            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.lock:
            self.calls.append(time.time())

            while self._timespan >= self.period:
                self.calls.popleft()

    @property
    def _timespan(self):
        return self.calls[-1] - self.calls[0]


class HTTPClient:
    def __init__(self):
        self.__limiter = _RateLimiter(10, 1)  # 10 reqs/seconds

    def request(self, method, url, **kwargs):
        with self.__limiter:
            r = requests.request(method=method, url=url, **kwargs)
        return r

    def fetch_image(self, url):
        headers = {
            'Accept': 'image/png,image/svg+xml,image/*;q=0.8,video/*;q=0.8,*/*;q=0.5',
            'Accept-Encoding': 'gzip, deflate, br',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) '
                          'Version/13.1.2 Safari/605.1.15',
            'Host': urllib.parse.urlparse(url).netloc, 'Accept-Language': 'en-ca', 'Referer': ROOT_URL,
            'Connection': 'keep-alive'
        }
        return self.request("GET", url, headers=headers, stream=True)


_default_http_client = HTTPClient()



from manganelo.common.constants import HOME_TOOLTIPS_URL
from manganelo.errors import RequestError
from manganelo.httpclient import _default_http_client
from manganelo.models import HomeStoryTooltip


def get_home_page() -> list[HomeStoryTooltip]:
    r = _default_http_client.request("GET", HOME_TOOLTIPS_URL)

    if r.status_code != 200:
        raise RequestError(f"Home page request failed with status code {r.status_code}")

    try:
        return [HomeStoryTooltip(ele) for ele in r.json()]
    except KeyError as e:
        raise RequestError("Home page response differs from expected response") from e
    except Exception as e:
        raise RequestError("An unexpected error occurred") from e



import datetime as dt

from pydantic import BaseModel

from manganelo.common import utils
from manganelo.httpclient import _default_http_client

from .chapter import Chapter
from .storypage import StoryPage


class SearchResult(BaseModel):
    title: str
    url: str
    icon_url: str
    authors: list[str]
    rating: float
    views: int
    updated: dt.datetime

    @property
    def story_page(self) -> StoryPage:
        from manganelo.storypage import get_story_page  # Circular import

        return get_story_page(self.url)

    @property
    def chapter_list(self) -> list[Chapter]:
        return self.story_page.chapter_list

    def download_icon(self, path: str):
        if img := _default_http_client.fetch_image(self.icon_url):
            return utils.save_image(img, path)

    @staticmethod
    def from_soup(soup):
        return SearchResult(
            title=_parse_title(soup),
            url=_parse_url(soup),
            icon_url=_parse_icon_url(soup),
            authors=_parse_authors(soup),
            rating=_parse_rating(soup),
            views=_parse_views(soup),
            updated=_parse_updated(soup)
        )


def _parse_title(soup):
    return soup.find(class_="item-img").get("title")


def _parse_url(soup):
    return soup.find(class_="item-img").get("href")


def _parse_icon_url(soup):
    return soup.find("img", class_="img-loading").get("src")


def _parse_authors(soup):
    authors = soup.find("span", class_="text-nowrap item-author")
    return utils.split_at(authors.text, ",") if authors else []


def _parse_rating(soup):
    return float(soup.find("em", class_="item-rate").text)


def _parse_views(soup):
    s = soup.find_all("span", class_="text-nowrap item-time")[-1].text
    number_string = s.replace("View : ", "").replace(",", "")
    return utils.parse_views(number_string)


def _parse_updated(soup):
    s = soup.find("span", class_="text-nowrap item-time").text
    return utils.parse_date(s, "Updated : %b %d,%Y - %H:%M")


import requests
from bs4 import BeautifulSoup

from manganelo.common import utils
from manganelo.common.constants import STORY_SEARCH_URL
from manganelo.errors import RequestError
from manganelo.models import SearchResult


def get_search_results(title: str) -> list[SearchResult]:
	r = requests.get(STORY_SEARCH_URL.format(title=utils.encode_querystring(title)))

	if r.status_code != 200:
		raise RequestError(f"Search request failed with status code {r.status_code}")

	soup = BeautifulSoup(r.content, "html.parser")

	return [SearchResult.from_soup(ele) for ele in soup.find_all(class_="search-story-item")]



import ast
import contextlib as cl
import datetime as dt
import html
import locale
import os
import shutil
import string
import typing


def parse_views(number_string: str) -> int:
	with cl.suppress(Exception):
		return ast.literal_eval(number_string)

	number_string, unit = number_string[:-1], number_string[-1].upper()

	multiplier = {
		"B": 1_000_000_000,
		"M": 1_000_000,
		"K": 1_000
	}.get(unit, 1)

	with cl.suppress(Exception):
		return int(float(number_string) * multiplier)

	return -1


def unescape_html(s: str) -> str:
	return html.unescape(s).strip()


def split_at(s: str, sep: str) -> list[str]:
	return [x.strip() for x in s.split(sep)]


def encode_querystring(s: str) -> str:
	allowed_characters: str = string.ascii_letters + string.digits + "_"

	return "".join([char.lower() for char in s.strip().replace(" ", "_") if char in allowed_characters])


def save_image(image_data, path) -> typing.Union[str, None]:
	path = validate_path(path)

	with open(path, "wb") as fh:
		image_data.raw.decode_content = True

		# noinspection PyBroadException
		try:
			shutil.copyfileobj(image_data.raw, fh)

		except BaseException:
			return None

	return path


def parse_date(s: str, _format: str):

	try:
		locale.setlocale(locale.LC_ALL, "en_US.UTF8")

		return dt.datetime.strptime(s, _format)

	finally:
		locale.setlocale(locale.LC_ALL, '')


def validate_path(path: str):
	dir_, file = os.path.split(path)

	for char in ("\\", "/", ":", "*", "?", "<", ">", "|"):
		file = file.replace(char, "")

	return os.path.join(dir_, file)



import os
import tempfile

from bs4 import BeautifulSoup
from PIL import Image
from reportlab.pdfgen import canvas

from manganelo.common import utils
from manganelo.httpclient import _default_http_client


def download_chapter(url, path):
	path = utils.validate_path(path)

	r = _default_http_client.request("GET", url)

	soup = BeautifulSoup(r.content, "html.parser")

	urls = _get_image_urls_from_soup(soup)

	with tempfile.TemporaryDirectory() as dir_:
		if images := _download_images(dir_, urls):
			_create_pdf(path, images)

	return path


def _get_image_urls_from_soup(soup):

	def valid(url: str):
		return url.endswith((".png", ".jpg"))

	return [url for url in map(lambda ele: ele["src"], soup.find_all("img")) if valid(url)]


def _download_images(dir_, urls: list):
	images = []

	for i, url in enumerate(urls):
		image = _default_http_client.fetch_image(url)

		if image is not None:
			ext = url.split(".")[-1]

			path = utils.save_image(image, os.path.join(dir_, f"{i}.{ext}"))

			if path:
				images.append(path)

	return images


def _create_pdf(path, images: list):
	pdf = canvas.Canvas(path)

	for image in images:

		# noinspection PyBroadException
		try:
			with Image.open(image) as img:
				w, h = img.size

		except BaseException:
			continue

		pdf.setPageSize((w, h))  # Set the page dimensions to the image dimensions

		pdf.drawImage(image, x=0, y=0)  # Insert the image onto the current page

		pdf.showPage()  # Create a new page ready for the next image

	pdf.save()



import datetime as dt

from manganelo.common import utils


class HomeStoryTooltip:
    __slots__ = ("id", "name", "other_names", "authors", "genres", "description", "update_time")

    def __init__(self, data: dict):
        self.id: int = data["id"]
        self.name: str = data["name"]
        self.other_names: list[str] = utils.split_at(data["nameother"], ";")
        self.authors: list[str] = utils.split_at(data["author"], ",")
        self.genres: list[str] = utils.split_at(data["genres"], ",")
        self.description: str = utils.unescape_html(data["description"])
        self.update_time: dt.datetime = utils.parse_date(data["updatetime"], "%b %d,%Y - %H:%M")



import datetime as dt

from pydantic import BaseModel

from manganelo.common import utils
from manganelo.httpclient import _default_http_client
from manganelo.models.chapter import Chapter


class StoryPage(BaseModel):
    url: str
    title: str
    icon_url: str
    description: str
    genres: list[str]
    views: int
    authors: list[str]
    updated: dt.datetime
    chapter_list: list[Chapter]

    def download_icon(self, path: str):
        if img := _default_http_client.fetch_image(self.icon_url):
            return utils.save_image(img, path)

    @staticmethod
    def from_soup(url: str, soup) -> "StoryPage":
        return StoryPage(
            url=url,
            title=_parse_title(soup),
            icon_url=_parse_icon_url(soup),
            description=_parse_description(soup),
            genres=_parse_genres(soup),
            views=_parse_views(soup),
            authors=_parse_authors(soup),
            updated=_parse_updated(soup),
            chapter_list=_parse_chapters(soup)
        )


def _parse_title(soup):
    return soup.find(class_="story-info-right").find("h1").text.strip()


def _parse_icon_url(soup):
    return soup.find("div", class_="story-info-left").find("img", class_="img-loading").get("src")


def _parse_description(soup):
    return utils.unescape_html(soup.find("div", class_="panel-story-info-description").text)


def _parse_authors(soup):
    authors_row = soup.find("i", class_="info-author").findNext("td", class_="table-value")
    return [e.strip() for e in authors_row.text.split(" - ")]


def _parse_genres(soup):
    genres_row = soup.find("i", class_="info-genres").findNext("td", class_="table-value")
    genres = genres_row.find_all("a", class_="a-h")
    return [e.text.strip() for e in genres]


def _parse_updated(soup) -> dt.datetime:
    values = soup.find("div", class_="story-info-right-extent").find_all("span", class_="stre-value")
    return utils.parse_date(values[0].text.strip(), "%b %d,%Y - %H:%M %p")


def _parse_views(soup) -> int:
    values = soup.find("div", class_="story-info-right-extent").find_all("span", class_="stre-value")
    s = values[1].text.strip().replace(",", "")
    return utils.parse_views(s)


def _parse_chapters(soup) -> list[Chapter]:
    panels = soup.find(class_="panel-story-chapter-list")
    return [Chapter.from_soup(ele) for ele in panels.find_all(class_="a-h")[::-1] if ele is not None]



from bs4 import BeautifulSoup

from manganelo.httpclient import _default_http_client

from .errors import RequestError
from .models import StoryPage


def get_story_page(url) -> StoryPage:
	r = _default_http_client.request("GET", url)

	soup = BeautifulSoup(r.content, "html.parser")

	if "404" in soup.find("title").text:
		raise RequestError(f"Page '{url}' was not found")

	return StoryPage.from_soup(url, soup)



from setuptools import find_packages, setup


def read_file(file):
	with open(file, "r") as fh:
		return fh.read()


setup(
	name="manganelo",
	packages=find_packages(),
	version="1.23",
	license="MIT",

	description="Unofficial API for the Manganelo/Manganato website.",
	long_description=read_file("README.md"),
	long_description_content_type="text/markdown",

	author="Joshua Nixon",
	author_email="nixonjoshua98@gmail.com",

	url="https://github.com/nixonjoshua98/manganelo",

	download_url="https://github.com/nixonjoshua98/manganelo/releases",

	keywords=[
		"manga", "manganelo", "scrapper", "web", "mangakakalot", "thread", "comic", "manhwa", "manganato", "manhwa"
	],

	install_requires=[
		"bs4",
		"requests",
		"reportlab",
		"pillow",
		"pydantic"
	],

	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"Topic :: Software Development :: Build Tools",
	],

	python_requires='>=3.9'
)